#include <metal_stdlib>
using namespace metal;

struct ModelConstants {
  float4x4 modelViewMatrix;
};

// Control Point struct
struct ControlPoint {
  float4 position [[attribute(0)]];
};

// Patch struct
struct PatchIn {
  patch_control_point<ControlPoint> control_points;
};

// Vertex-to-Fragment struct
struct VertexOut {
  float4 position [[position]];
  half4  color [[flat]];
};

// Quad compute kernel
kernel void tessellation_kernel_quad(constant float& edge_factor [[ buffer(0) ]],
                                     constant float& inside_factor [[ buffer(1) ]],
                                     device MTLQuadTessellationFactorsHalf* factors [[ buffer(2) ]],
                                     uint pid [[ thread_position_in_grid ]])
{
  factors[pid].edgeTessellationFactor[0] = edge_factor;
  factors[pid].edgeTessellationFactor[1] = edge_factor;
  factors[pid].edgeTessellationFactor[2] = edge_factor;
  factors[pid].edgeTessellationFactor[3] = edge_factor;
  factors[pid].insideTessellationFactor[0] = inside_factor;
  factors[pid].insideTessellationFactor[1] = inside_factor;
}


// Quad post-tessellation vertex function
// Biquadratic patch with nine control points
[[patch(quad, 16)]]
vertex VertexOut tessellation_vertex_quad(PatchIn patchIn [[stage_in]],
                                              float2 patch_coord [[ position_in_patch ]],
                                              constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;

  float u = patch_coord.x;
  float v = patch_coord.y;
  
  float3 p00 = patchIn.control_points[0].position.xyz;
  float3 p01 = patchIn.control_points[1].position.xyz;
  float3 p02 = patchIn.control_points[2].position.xyz;
  float3 p03 = patchIn.control_points[3].position.xyz;
  float3 p10 = patchIn.control_points[4].position.xyz;
  float3 p11 = patchIn.control_points[5].position.xyz;
  float3 p12 = patchIn.control_points[6].position.xyz;
  float3 p13 = patchIn.control_points[7].position.xyz;
  float3 p20 = patchIn.control_points[8].position.xyz;
  float3 p21 = patchIn.control_points[9].position.xyz;
  float3 p22 = patchIn.control_points[10].position.xyz;
  float3 p23 = patchIn.control_points[11].position.xyz;
  float3 p30 = patchIn.control_points[12].position.xyz;
  float3 p31 = patchIn.control_points[13].position.xyz;
  float3 p32 = patchIn.control_points[14].position.xyz;
  float3 p33 = patchIn.control_points[15].position.xyz;

  float4x4 px;
  px[0] = float4(p00.x, p10.x, p20.x, p30.x);
  px[1] = float4(p01.x, p11.x, p21.x, p31.x);
  px[2] = float4(p02.x, p12.x, p22.x, p32.x);
  px[3] = float4(p03.x, p13.x, p23.x, p33.x);

  float4x4 py;
  py[0] = float4(p00.y, p10.y, p20.y, p30.y);
  py[1] = float4(p01.y, p11.y, p21.y, p31.y);
  py[2] = float4(p02.y, p12.y, p22.y, p32.y);
  py[3] = float4(p03.y, p13.y, p23.y, p33.y);

  float4x4 pz;
  pz[0] = float4(p00.z, p10.z, p20.z, p30.z);
  pz[1] = float4(p01.z, p11.z, p21.z, p31.z);
  pz[2] = float4(p02.z, p12.z, p22.z, p32.z);
  pz[3] = float4(p03.z, p13.z, p23.z, p33.z);

  float4x4 M;
  M[0] = float4(1, -3,  3,  1);
  M[1] = float4(0,  3, -6,  3);
  M[2] = float4(0,  0,  3, -3);
  M[3] = float4(0,  0,  0,  1);
  
  // transpose of M
  float4x4 MT;
  MT[0] = float4(1, 0, 0, 0);
  MT[1] = float4(-3, 3, 0, 0);
  MT[2] = float4(3, -6, 3, 0);
  MT[3] = float4(1, 3, -3, 1);
  
  // bicubic interpolation
  // 3D Graphics for Game Programming p.150
  float usquared = u * u;
  float vsquared = v * v;
  float ucubed = usquared * u;
  float vcubed = vsquared * v;
  
  float uinverse = 1-u;
  float vinverse = 1-v;
  float uinvsquared = uinverse * uinverse;
  float vinvsquared = vinverse * vinverse;
  float uinvcubed = uinvsquared * uinverse;
  float vinvcubed = vinvsquared * vinverse;
  
  /*
   //http://twvideo01.ubm-us.net/o1/vault/gdc10/slides/Bilodeau_Bill_Direct3D11TutorialTessellation.pdf p24
   */
  
  float invT = uinverse;
  float4 U = float4(uinvcubed,
                         3 * u * uinvsquared,
                         3 * usquared * uinverse,
                         ucubed);
  invT = 1.0 - v;
  float4 V = float4(vinvcubed,
                         3 * v * vinvsquared,
                         3 * vsquared * vinverse,
                         vcubed);

  float3 uv  = V.x * (p00 * U.x + p01 * U.y + p02 * U.z + p03 * U.w);
  uv += V.y * (p10 * U.x + p11 * U.y + p12 * U.z + p13 * U.w);
  uv += V.z * (p20 * U.x + p21 * U.y + p22 * U.z + p23 * U.w);
  uv += V.w * (p30 * U.x + p31 * U.y + p32 * U.z + p33 * U.w);

  vertexOut.position = float4(uv, 1);
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  vertexOut.color = half4(u, v, 1.0-v, 1.0);
  return vertexOut;
}

fragment half4 tessellation_fragment(VertexOut fragmentIn [[stage_in]])
{
  return fragmentIn.color;
}


