
import MetalKit
import simd

struct ModelConstants {
  var modelViewMatrix = matrix_identity_float4x4
}

let numberOfPatchControlPoints = 16

public class Renderer: NSObject {
  let device: MTLDevice
  let commandQueue: MTLCommandQueue
  var library: MTLLibrary?

  // controls tessellation amount
  public var edgeFactor: [Float] = [8.0]
  public var insideFactor: [Float] = [8.0]

  var computePipelineState: MTLComputePipelineState!
  var tessellationRenderPipelineState: MTLRenderPipelineState?
  var controlPointsBuffer: MTLBuffer!
  var tessellationFactorsBuffer: MTLBuffer!

  var depthStencil: MTLDepthStencilState!
  
  var position = float3(repeating: 0)
  var rotation = float3(repeating: 0)
  var scale = float3(repeating: 1)
  
  var modelMatrix: float4x4 {
    let translationMatrix = float4x4(translation: position)
    let rotationMatrix = float4x4(rotation: rotation)
    let scaleMatrix = float4x4(scaling: scale)
    let matrix = translationMatrix * scaleMatrix * rotationMatrix
    return matrix
  }

  public init(device: MTLDevice) {
    self.device = device
    commandQueue = device.makeCommandQueue()!
    
//    do {
//      let path = Bundle.main.path(forResource: "Tessellation", ofType: "metal")
//      let source = try String(contentsOfFile: path!, encoding: .utf8)
//      library = try device.makeLibrary(source: source, options: nil)
//    } catch let error as NSError {
//      print("library error: " + error.description)
//    }

    library = device.makeDefaultLibrary()!

    super.init()
    setupTessellationPipeline()
    setupDepthStencil()
    
  }

  func setupTessellationPipeline() {
    
    // Setup Compute Pipeline
    let kernelFunction = library?.makeFunction(name: "tessellation_kernel_quad")
    do {
      computePipelineState = try device.makeComputePipelineState(function: kernelFunction!)
    } catch let error as NSError {
      print("compute pipeline error: " + error.description)
    }
    
    // Setup Vertex Descriptor
    let vertexDescriptor = MTLVertexDescriptor()
    vertexDescriptor.attributes[0].format = .float4;
    vertexDescriptor.attributes[0].offset = 0;
    vertexDescriptor.attributes[0].bufferIndex = 0;
    vertexDescriptor.layouts[0].stepFunction = .perPatchControlPoint;
    vertexDescriptor.layouts[0].stepRate = 1;
    vertexDescriptor.layouts[0].stride = 4*MemoryLayout<Float>.size;
    
    // Setup Render Pipeline
    let tessDescriptor = MTLRenderPipelineDescriptor()
    tessDescriptor.vertexDescriptor = vertexDescriptor
    tessDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
    tessDescriptor.depthAttachmentPixelFormat = .depth32Float
    
    tessDescriptor.fragmentFunction = library?.makeFunction(name: "tessellation_fragment")
    tessDescriptor.isTessellationFactorScaleEnabled = false
    tessDescriptor.tessellationFactorFormat = .half
    tessDescriptor.tessellationControlPointIndexType = .none
    tessDescriptor.tessellationFactorStepFunction = .constant
    tessDescriptor.tessellationOutputWindingOrder = .clockwise
    tessDescriptor.tessellationPartitionMode = .fractionalEven
    tessDescriptor.maxTessellationFactor = 64;
    tessDescriptor.vertexFunction = library?.makeFunction(name: "tessellation_vertex_quad")
    do {
      tessellationRenderPipelineState = try device.makeRenderPipelineState(descriptor: tessDescriptor)
    } catch let error as NSError {
      print("render pipeline error: " + error.description)
      
    }
    
    // Set up buffers
    tessellationFactorsBuffer = device.makeBuffer(length: 256, options: MTLResourceOptions.storageModePrivate)
    
    controlPointsBuffer = device.makeBuffer(bytes: controlPoints, length:256 , options: [])
  }

  func setupDepthStencil() {
    let depthStateDesc = MTLDepthStencilDescriptor()
    depthStateDesc.depthCompareFunction = .less // seems to do nothing
    depthStateDesc.isDepthWriteEnabled = true
    depthStencil = device.makeDepthStencilState(descriptor: depthStateDesc)
  }

  public func rotateUsing(translation: NSPoint) {
    let sensitivity: Float = 0.01
    rotation.x += Float(translation.y) * sensitivity
    rotation.z -= Float(translation.x) * sensitivity
  }
}

extension Renderer: MTKViewDelegate {
  public func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
  }
  
  
  public func draw(in view: MTKView) {
    guard let drawable = view.currentDrawable,
      let descriptor = view.currentRenderPassDescriptor else { return }
    
    let commandBuffer = commandQueue.makeCommandBuffer()!

    var modelConstants = ModelConstants()
    
    let viewMatrix = float4x4(translation: [0, 0, -2])
    let aspect: Float = 1.0
    let projectionMatrix = float4x4(projectionFov: radians(fromDegrees: 65),
                                    near: 0.1, far: 100, aspect: aspect)

    let modelViewMatrix = viewMatrix * modelMatrix
    modelConstants.modelViewMatrix = projectionMatrix * modelViewMatrix
    
    
    let computeCommandEncoder = commandBuffer.makeComputeCommandEncoder()!
    computeCommandEncoder.setComputePipelineState(computePipelineState!)
    computeCommandEncoder.setBytes(edgeFactor, length: MemoryLayout<Float>.size, index: 0)
    computeCommandEncoder.setBytes(insideFactor, length: MemoryLayout<Float>.size, index: 1)
    computeCommandEncoder.setBuffer(tessellationFactorsBuffer, offset: 0, index: 2)
    computeCommandEncoder.dispatchThreadgroups(MTLSizeMake(1, 1, 1), threadsPerThreadgroup: MTLSizeMake(1, 1, 1))
    computeCommandEncoder.endEncoding()
    
    let renderCommandEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: descriptor)!
    renderCommandEncoder.setRenderPipelineState(tessellationRenderPipelineState!)
    renderCommandEncoder.setVertexBuffer(controlPointsBuffer, offset: 0, index: 0)
    
    renderCommandEncoder.setVertexBytes(&modelConstants,
                                        length: MemoryLayout<ModelConstants>.stride, index: 1)
    
    renderCommandEncoder.setTriangleFillMode(.lines)
    renderCommandEncoder.setTessellationFactorBuffer(tessellationFactorsBuffer, offset: 0, instanceStride: 0)
    let patchCount = controlPoints.count / 4 / 4
    renderCommandEncoder.drawPatches(numberOfPatchControlPoints: numberOfPatchControlPoints,
                                     patchStart: 0, patchCount: patchCount,
                                     patchIndexBuffer: nil, patchIndexBufferOffset: 0,
                                     instanceCount: 1, baseInstance: 0)
    renderCommandEncoder.endEncoding()
    commandBuffer.present(drawable)
    commandBuffer.commit()
    commandBuffer.waitUntilCompleted()
    
  }

}



