#include <metal_stdlib>
using namespace metal;

struct ModelConstants {
  float4x4 modelViewMatrix;
};

// Control Point struct
struct ControlPoint {
  float4 position [[attribute(0)]];
};

// Patch struct
struct PatchIn {
  patch_control_point<ControlPoint> control_points;
};

// Vertex-to-Fragment struct
struct VertexOut {
  float4 position [[position]];
  half4  color [[flat]];
};

// Quad compute kernel
kernel void tessellation_kernel_quad(constant float& edge_factor [[ buffer(0) ]],
                                     constant float& inside_factor [[ buffer(1) ]],
                                     device MTLQuadTessellationFactorsHalf* factors [[ buffer(2) ]],
                                     uint pid [[ thread_position_in_grid ]])
{
  factors[pid].edgeTessellationFactor[0] = edge_factor;
  factors[pid].edgeTessellationFactor[1] = edge_factor;
  factors[pid].edgeTessellationFactor[2] = edge_factor;
  factors[pid].edgeTessellationFactor[3] = edge_factor;
  factors[pid].insideTessellationFactor[0] = inside_factor;
  factors[pid].insideTessellationFactor[1] = inside_factor;
}


// Quad post-tessellation vertex function
// Biquadratic patch with nine control points
[[patch(quad, 9)]]
vertex VertexOut tessellation_vertex_quad(PatchIn patchIn [[stage_in]],
                                              float2 patch_coord [[ position_in_patch ]],
                                              constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;

  float u = patch_coord.x;
  float v = patch_coord.y;
  
  float3 p00 = patchIn.control_points[0].position.xyz;
  float3 p01 = patchIn.control_points[1].position.xyz;
  float3 p02 = patchIn.control_points[2].position.xyz;
  float3 p10 = patchIn.control_points[3].position.xyz;
  float3 p11 = patchIn.control_points[4].position.xyz;
  float3 p12 = patchIn.control_points[5].position.xyz;
  float3 p20 = patchIn.control_points[6].position.xyz;
  float3 p21 = patchIn.control_points[7].position.xyz;
  float3 p22 = patchIn.control_points[8].position.xyz;
  
  // biquadratic interpolation
  // 3D Graphics for Game Programming p.146
  float usquared = u * u;
  float vsquared = v * v;
  float uinverse = 1-u;
  float vinverse = 1-v;
  float uinvsquared = uinverse * uinverse;
  float vinvsquared = vinverse * vinverse;
  float3 uv = uinvsquared * vinvsquared * p00 + 2 * u * uinverse * vinvsquared * p01 + usquared * vinvsquared * p02 + 2 * uinvsquared * v * vinverse * p10 + 4 * u * v * uinverse * vinverse * p11 + 2 * usquared * v * vinverse * p12 + uinvsquared * vsquared * p20 + 2 * u * uinverse * vsquared * p21 + usquared * vsquared * p22;
  
  vertexOut.position = float4(uv, 1);
  vertexOut.position.x = u;
  vertexOut.position.y = v;
  vertexOut.position.z = 0;
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  vertexOut.color = half4(u, v, 1.0-v, 1.0);
  return vertexOut;
}

fragment half4 tessellation_fragment(VertexOut fragmentIn [[stage_in]])
{
  return fragmentIn.color;
}


