import MetalKit
import simd

struct ModelConstants {
  var modelViewMatrix = matrix_identity_float4x4
}

public class Renderer: NSObject {
  let device: MTLDevice
  let commandQueue: MTLCommandQueue
  var library: MTLLibrary?
  
  var depthStencil: MTLDepthStencilState!
  
  
  var position = float3(repeating: 0)
  var rotation = float3(repeating: 0)
  var scale = float3(repeating: 1)
  
  var modelMatrix: float4x4 {
    let translationMatrix = float4x4(translation: position)
    let rotationMatrix = float4x4(rotation: rotation)
    let scaleMatrix = float4x4(scaling: scale)
    let matrix = translationMatrix * scaleMatrix * rotationMatrix
    return matrix
  }

  let vertexBuffer: MTLBuffer
  let indexBuffer: MTLBuffer
  var pipelineState: MTLRenderPipelineState!

  var vertexDescriptor: MTLVertexDescriptor {
    let vertexDescriptor = MTLVertexDescriptor()
    
    vertexDescriptor.attributes[0].format = .float3
    vertexDescriptor.attributes[0].offset = 0
    vertexDescriptor.attributes[0].bufferIndex = 0
    
    vertexDescriptor.attributes[1].format = .float4
    vertexDescriptor.attributes[1].offset = MemoryLayout<float3>.stride
    vertexDescriptor.attributes[1].bufferIndex = 0
    
    vertexDescriptor.attributes[2].format = .float2
    vertexDescriptor.attributes[2].offset = MemoryLayout<float3>.stride + MemoryLayout<float4>.stride
    vertexDescriptor.attributes[2].bufferIndex = 0
    
    vertexDescriptor.layouts[0].stride = MemoryLayout<Vertex>.stride
    
    return vertexDescriptor
  }
  

  
  public init(device: MTLDevice) {
    self.device = device
    commandQueue = device.makeCommandQueue()!

    vertices[0].color = float4(1.0, 0.0, 0.0, 1.0)
    vertices[1].color = float4(0, 1, 0, 1)
    vertices[2].color = float4(0, 0, 1, 1)
    vertices[3].color = float4(1, 0, 1, 1)
    vertices[4].color = float4(0, 0, 1, 1)
    vertices[5].color = float4(0, 1, 0, 1)
    vertices[6].color = float4(1, 0, 0, 1)
    vertices[7].color = float4(1, 0, 1, 1)
    
    vertexBuffer = device.makeBuffer(bytes: vertices,
                                     length: vertices.count *
                                      MemoryLayout<Vertex>.stride,
                                     options: [])!
    indexBuffer = device.makeBuffer(bytes: indices,
                                    length: indices.count * MemoryLayout<UInt16>.size,
                                    options: [])!
    print("here1")

//    do {
//      let path = Bundle.main.path(forResource: "Shader", ofType: "metal")
//      let source = try String(contentsOfFile: path!, encoding: .utf8)
//      library = try device.makeLibrary(source: source, options: nil)
//    } catch let error as NSError {
//      print("library error: " + error.description)
//    }
    print("here2")
    library = device.makeDefaultLibrary()!

    super.init()
    print("here3")
    setupPipeline()
    setupDepthStencil()
    
  }

  func setupPipeline() {
    let vertexFunction = library?.makeFunction(name: "vertex_shader")
    let fragmentFunction = library?.makeFunction(name: "fragment_shader")
    
    let pipelineDescriptor = MTLRenderPipelineDescriptor()
    pipelineDescriptor.vertexFunction = vertexFunction
    pipelineDescriptor.fragmentFunction = fragmentFunction
    pipelineDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
    pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
    pipelineDescriptor.vertexDescriptor = vertexDescriptor
    do {
      pipelineState = try device.makeRenderPipelineState(descriptor: pipelineDescriptor)
    } catch let error as NSError {
      fatalError("error: \(error.localizedDescription)")
    }
  }
  
  func setupDepthStencil() {
    let depthStateDesc = MTLDepthStencilDescriptor()
    depthStateDesc.depthCompareFunction = .less // seems to do nothing
    depthStateDesc.isDepthWriteEnabled = true
    depthStencil = device.makeDepthStencilState(descriptor: depthStateDesc)
  }

  public func rotateUsing(translation: NSPoint) {
    let sensitivity: Float = 0.01
    rotation.x += Float(translation.y) * sensitivity
    rotation.z -= Float(translation.x) * sensitivity
  }
}

extension Renderer: MTKViewDelegate {
  public func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
  }
  
  
  public func draw(in view: MTKView) {
    guard let drawable = view.currentDrawable,
      let descriptor = view.currentRenderPassDescriptor else { return }
    
    let commandBuffer = commandQueue.makeCommandBuffer()!
    
    var modelConstants = ModelConstants()
    
    let viewMatrix = float4x4(translation: [0, 0, -2])
    let aspect: Float = 1.0
    let projectionMatrix = float4x4(projectionFov: radians(fromDegrees: 65),
                                    near: 0.1, far: 100, aspect: aspect)
    
    let modelViewMatrix = viewMatrix * modelMatrix
    modelConstants.modelViewMatrix = projectionMatrix * modelViewMatrix
    
    // Render
    
    let commandEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: descriptor)!
    commandEncoder.setRenderPipelineState(pipelineState)
    commandEncoder.setDepthStencilState(depthStencil)
    commandEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: 0)
    commandEncoder.setVertexBytes(&modelConstants,
                                  length: MemoryLayout<ModelConstants>.stride, index: 1)
    
    commandEncoder.setTriangleFillMode(.lines) 
    
    commandEncoder.drawIndexedPrimitives(type: .triangle,
                                         indexCount: indices.count,
                                         indexType: .uint16,
                                         indexBuffer: indexBuffer,
                                         indexBufferOffset: 0)
    
    commandEncoder.endEncoding()
    
    commandBuffer.present(drawable)
    commandBuffer.commit()
    
  }
  
}



