#include <metal_stdlib>
using namespace metal;

struct ModelConstants {
  float4x4 modelViewMatrix;
};

// Control Point struct
struct ControlPoint {
  float4 position [[attribute(0)]];
};

// Patch struct
struct PatchIn {
  patch_control_point<ControlPoint> control_points;
};

// Vertex-to-Fragment struct
struct VertexOut {
  float4 position [[position]];
  half4  color [[flat]];
};

// Quad compute kernel
kernel void tessellation_kernel_quad(constant float& edge_factor [[ buffer(0) ]],
                                     constant float& inside_factor [[ buffer(1) ]],
                                     device MTLQuadTessellationFactorsHalf* factors [[ buffer(2) ]],
                                     uint pid [[ thread_position_in_grid ]])
{
  factors[pid].edgeTessellationFactor[0] = edge_factor;
  factors[pid].edgeTessellationFactor[1] = edge_factor;
  factors[pid].edgeTessellationFactor[2] = edge_factor;
  factors[pid].edgeTessellationFactor[3] = edge_factor;
  factors[pid].insideTessellationFactor[0] = inside_factor;
  factors[pid].insideTessellationFactor[1] = inside_factor;
}


// Quad post-tessellation vertex function
// Bilinear patch with four control points
[[patch(quad, 4)]]
vertex VertexOut tessellation_vertex_quad(PatchIn patchIn [[stage_in]],
                                              float2 patch_coord [[ position_in_patch ]],
                                              constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;

  float u = patch_coord.x;
  float v = patch_coord.y;
  
  float3 p00 = patchIn.control_points[1].position.xyz;
  float3 p01 = patchIn.control_points[2].position.xyz;
  float3 p10 = patchIn.control_points[0].position.xyz;
  float3 p11 = patchIn.control_points[3].position.xyz;
  
  // bilinear interpolation
  // 3D Graphics for Game Programming p.141
  float3 uv = (1-u) * (1-v) * p00 + u * (1 - v) * p01 + (1 - u) * v * p10 + u * v * p11;
  
  // better interpolation:
  // http://reedbeta.com/blog/tess-quick-ref/
  
  uv = mix(mix(p00, p01, u),
           mix(p10, p11, u),
           v);
  
  vertexOut.position = float4(uv, 1);
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  vertexOut.color = half4(u, v, 1.0-v, 1.0);
  return vertexOut;
}

fragment half4 tessellation_fragment(VertexOut fragmentIn [[stage_in]])
{
  return fragmentIn.color;
}


